@@ -46,51 +46,51 @@ module Distribution.Extra.Doctest (
 
 import Control.Monad
        (when)
+import Data.IORef
+       (modifyIORef, newIORef, readIORef)
 import Data.List
        (nub)
 import Data.Maybe
-       (maybeToList, mapMaybe)
+       (mapMaybe, maybeToList)
 import Data.String
        (fromString)
-import qualified Data.Foldable as F
-       (for_)
-import qualified Data.Traversable as T
-       (traverse)
-import qualified Distribution.ModuleName as ModuleName
-       (fromString)
-import Distribution.ModuleName
-       (ModuleName)
 import Distribution.Package
-       (InstalledPackageId)
-import Distribution.Package
-       (Package (..), PackageId, packageVersion)
+       (InstalledPackageId, Package (..))
 import Distribution.PackageDescription
-       (BuildInfo (..), Executable (..), Library (..), GenericPackageDescription,
-       PackageDescription (), TestSuite (..))
+       (BuildInfo (..), Executable (..), GenericPackageDescription,
+       Library (..), PackageDescription, TestSuite (..))
 import Distribution.Simple
-       (UserHooks (..), autoconfUserHooks, defaultMainWithHooks, simpleUserHooks)
-import Distribution.Simple.BuildPaths
-       (autogenModulesDir)
+       (UserHooks (..), autoconfUserHooks, defaultMainWithHooks,
+       simpleUserHooks)
 import Distribution.Simple.Compiler
-       (PackageDB (..), showCompilerId)
+       (CompilerFlavor (GHC), CompilerId (..), PackageDB (..), compilerId)
 import Distribution.Simple.LocalBuildInfo
-       (ComponentLocalBuildInfo (componentPackageDeps), LocalBuildInfo (),
+       (ComponentLocalBuildInfo (componentPackageDeps), LocalBuildInfo,
        compiler, withExeLBI, withLibLBI, withPackageDB, withTestLBI)
 import Distribution.Simple.Setup
-       (BuildFlags (buildDistPref, buildVerbosity), HaddockFlags (haddockDistPref, haddockVerbosity), fromFlag, emptyBuildFlags)
+       (BuildFlags (buildDistPref, buildVerbosity),
+       HaddockFlags (haddockDistPref, haddockVerbosity), emptyBuildFlags,
+       fromFlag)
 import Distribution.Simple.Utils
-       (createDirectoryIfMissingVerbose, findFile, rewriteFile)
+       (createDirectoryIfMissingVerbose, info)
 import Distribution.Text
-       (display, simpleParse)
+       (display)
 import System.FilePath
-       ((</>), (<.>), dropExtension)
+       ((</>))
 
-import Data.IORef (newIORef, modifyIORef, readIORef)
+import qualified Data.Foldable    as F
+                 (for_)
+import qualified Data.Traversable as T
+                 (traverse)
 
 #if MIN_VERSION_Cabal(1,25,0)
 import Distribution.Simple.BuildPaths
        (autogenComponentModulesDir)
+#else
+import Distribution.Simple.BuildPaths
+       (autogenModulesDir)
 #endif
+
 #if MIN_VERSION_Cabal(2,0,0)
 import Distribution.Types.MungedPackageId
        (MungedPackageId)
@@ -98,10 +98,31 @@ import Distribution.Types.UnqualComponentName
        (unUnqualComponentName)
 
 -- For amendGPD
-import Distribution.Types.GenericPackageDescription
-       (GenericPackageDescription (condTestSuites))
 import Distribution.PackageDescription
        (CondTree (..))
+import Distribution.Types.GenericPackageDescription
+       (GenericPackageDescription (condTestSuites))
+
+import Distribution.Version
+       (mkVersion)
+#else
+import Data.Version
+       (Version (..))
+import Distribution.Package
+       (PackageId)
+#endif
+
+#if MIN_VERSION_Cabal(3,0,0)
+import Distribution.Simple.Utils
+       (findFileEx)
+#else
+import Distribution.Simple.Utils
+       (findFile)
+#endif
+
+#if MIN_VERSION_Cabal(3,0,0)
+import Distribution.Types.LibraryName
+       (libraryNameString)
 #endif
 
 #if MIN_VERSION_directory(1,2,2)
@@ -120,6 +141,20 @@ makeAbsolute p | isAbsolute p = return p
     return $ cwd </> p
 #endif
 
+#if !MIN_VERSION_Cabal(3,0,0)
+findFileEx :: verbosity -> [FilePath] -> FilePath -> IO FilePath
+findFileEx _ = findFile
+#endif
+
+#if !MIN_VERSION_Cabal(2,0,0)
+mkVersion :: [Int] -> Version
+mkVersion ds = Version ds []
+#endif
+
+-------------------------------------------------------------------------------
+-- Mains
+-------------------------------------------------------------------------------
+
 -- | A default main with doctests:
 --
 -- @
@@ -216,9 +251,12 @@ generateBuildModule testSuiteName flags pkg lbi = do
   let verbosity = fromFlag (buildVerbosity flags)
   let distPref = fromFlag (buildDistPref flags)
 
-  -- Package DBs
+  -- Package DBs & environments
   let dbStack = withPackageDB lbi ++ [ SpecificPackageDB $ distPref </> "package.conf.inplace" ]
   let dbFlags = "-hide-all-packages" : packageDbArgs dbStack
+  let envFlags
+        | ghcCanBeToldToIgnorePkgEnvs = [ "-package-env=-" ]
+        | otherwise = []
 
   withTestLBI pkg lbi $ \suite suitecfg -> when (testName suite == fromString testSuiteName) $ do
 #if MIN_VERSION_Cabal(1,25,0)
@@ -233,6 +271,7 @@ generateBuildModule testSuiteName flags pkg lbi = do
 
     -- First, we create the autogen'd module Build_doctests.
     -- Initially populate Build_doctests with a simple preamble.
+    info verbosity $ "cabal-doctest: writing Build_doctests to " ++ buildDoctestsFile
     writeFile buildDoctestsFile $ unlines
       [ "module Build_doctests where"
       , ""
@@ -306,7 +345,7 @@ generateBuildModule testSuiteName flags pkg lbi = do
            -- even though the main-is module is named Main, its filepath might
            -- actually be Something.hs. To account for this possibility, we simply
            -- pass the full path to the main-is module instead.
-           mainIsPath <- T.traverse (findFile iArgsNoPrefix) (compMainIs comp)
+           mainIsPath <- T.traverse (findFileEx verbosity iArgsNoPrefix) (compMainIs comp)
 
            let all_sources = map display module_sources
                              ++ additionalModules
@@ -319,6 +358,7 @@ generateBuildModule testSuiteName flags pkg lbi = do
                   [ iArgs
                   , additionalDirs
                   , includeArgs
+                  , envFlags
                   , dbFlags
                   , cppFlags
                   , extensionArgs
@@ -334,8 +374,8 @@ generateBuildModule testSuiteName flags pkg lbi = do
     getBuildDoctests withExeLBI (NameExe . executableName) (const [])     (Just . modulePath) buildInfo
 
     components <- readIORef componentsRef
-    F.for_ components $ \(Component name pkgs flags sources) -> do
-       let compSuffix          = nameToString name
+    F.for_ components $ \(Component cmpName cmpPkgs cmpFlags cmpSources) -> do
+       let compSuffix          = nameToString cmpName
            pkgs_comp           = "pkgs"           ++ compSuffix
            flags_comp          = "flags"          ++ compSuffix
            module_sources_comp = "module_sources" ++ compSuffix
@@ -344,13 +384,13 @@ generateBuildModule testSuiteName flags pkg lbi = do
        appendFile buildDoctestsFile $ unlines
          [ -- -package-id etc. flags
            pkgs_comp ++ " :: [String]"
-         , pkgs_comp ++ " = " ++ show pkgs
+         , pkgs_comp ++ " = " ++ show cmpPkgs
          , ""
          , flags_comp ++ " :: [String]"
-         , flags_comp ++ " = " ++ show flags
+         , flags_comp ++ " = " ++ show cmpFlags
          , ""
          , module_sources_comp ++ " :: [String]"
-         , module_sources_comp ++ " = " ++ show sources
+         , module_sources_comp ++ " = " ++ show cmpSources
          , ""
          ]
 
@@ -376,10 +416,14 @@ generateBuildModule testSuiteName flags pkg lbi = do
     parseComponentName _ = Nothing
 
     -- we do this check in Setup, as then doctests don't need to depend on Cabal
-    isOldCompiler = maybe False id $ do
-      a <- simpleParse $ showCompilerId $ compiler lbi
-      b <- simpleParse "7.5"
-      return $ packageVersion (a :: PackageId) < b
+    isNewCompiler = case compilerId $ compiler lbi of
+      CompilerId GHC v -> v >= mkVersion [7,6]
+      _                -> False
+
+    ghcCanBeToldToIgnorePkgEnvs :: Bool
+    ghcCanBeToldToIgnorePkgEnvs = case compilerId $ compiler lbi of
+      CompilerId GHC v -> v >= mkVersion [8,4,4]
+      _                -> False
 
     formatDeps = map formatOne
     formatOne (installedPkgId, pkgId)
@@ -396,8 +440,8 @@ generateBuildModule testSuiteName flags pkg lbi = do
 
     -- From Distribution.Simple.Program.GHC
     packageDbArgs :: [PackageDB] -> [String]
-    packageDbArgs | isOldCompiler = packageDbArgsConf
-                  | otherwise     = packageDbArgsDb
+    packageDbArgs | isNewCompiler = packageDbArgsDb
+                  | otherwise     = packageDbArgsConf
 
     -- GHC <7.6 uses '-package-conf' instead of '-package-db'.
     packageDbArgsConf :: [PackageDB] -> [String]
@@ -432,7 +476,9 @@ generateBuildModule testSuiteName flags pkg lbi = do
        isSpecific _                     = False
 
     mbLibraryName :: Library -> Name
-#if MIN_VERSION_Cabal(2,0,0)
+#if MIN_VERSION_Cabal(3,0,0)
+    mbLibraryName = NameLib . fmap unUnqualComponentName . libraryNameString . libName
+#elif MIN_VERSION_Cabal(2,0,0)
     -- Cabal-2.0 introduced internal libraries, which are named.
     mbLibraryName = NameLib . fmap unUnqualComponentName . libName
 #else
@@ -462,7 +508,7 @@ amendGPD
     -> GenericPackageDescription
     -> GenericPackageDescription
 #if !(MIN_VERSION_Cabal(2,0,0))
-amendGPD _ = id
+amendGPD _ gpd = gpd
 #else
 amendGPD testSuiteName gpd = gpd
     { condTestSuites = map f (condTestSuites gpd)
